#!/usr/bin/env python
# encoding: utf-8
"""
*The recipe to generate a master dark frame*

:Author:
    David Young & Marco Landoni

:Date Created:
    January 27, 2020
"""
################# GLOBAL IMPORTS ####################
from builtins import object
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools
from soxspipe.commonutils import set_of_files
from ._base_recipe_ import _base_recipe_
from astropy.stats import mad_std
import numpy as np
from astropy.nddata import CCDData
import ccdproc


class mdark(_base_recipe_):
    """
    *The mdark recipe*

    **Key Arguments:**
        - ``log`` -- logger
        - ``settings`` -- the settings dictionary
        - ``inputFrames`` -- input fits frames. Can be a directory, a set-of-files (SOF) file or a list of fits frame paths. Default []

    **Usage:**

    To setup your logger, settings and database connections, please use the ``fundamentals`` package (`see tutorial here <http://fundamentals.readthedocs.io/en/latest/#tutorial>`_). 

    See `produce_product` method for usage.

    ```eval_rst
    .. todo::

        - add a tutorial about ``mdark`` to documentation

    ```
    """
    # Initialisation

    def __init__(
            self,
            log,
            settings=False,
            inputFrames=[]

    ):
        # INHERIT INITIALISATION FROM  _base_recipe_
        super(mdark, self).__init__(log=log, settings=settings)
        self.log = log
        log.debug("instansiating a new 'mdark' object")
        self.settings = settings
        self.inputFrames = inputFrames
        # xt-self-arg-tmpx

        # INITIAL ACTIONS
        # CONVERT INPUT FILES TO A CCDPROC IMAGE COLLECTION (inputFrames >
        # imagefilecollection)
        sof = set_of_files(
            log=self.log,
            settings=self.settings,
            inputFrames=self.inputFrames
        )
        self.inputFrames = sof.get()

        # VERIFY THE FRAMES ARE THE ONES EXPECTED BY MDARK - NO MORE, NO LESS.
        # PRINT SUMMARY OF FILES.
        self.verify_input_frames()
        print(self.inputFrames.summary)

        # PREPARE THE FRAMES - CONVERT TO ELECTRONS, ADD UNCERTAINTY AND MASK
        # EXTENSIONS
        self.inputFrames = self.prepare_frames(
            save=self.settings["save-intermediate-products"])

        return None

    def verify_input_frames(
            self):
        """*verify the input frame match those required by the mdark recipe*

        **Return:**
            - ``None``

        If the fits files conform to required input for the recipe everything will pass silently, otherwise an exception shall be raised.
        """
        self.log.debug('starting the ``verify_input_frames`` method')

        imageTypes = self.inputFrames.values(
            keyword='eso dpr type', unique=True)

        # MIXED INPUT IMAGE TYPES ARE BAD
        if len(imageTypes) > 1:
            imageTypes = " and ".join(imageTypes)
            print(self.inputFrames.summary)
            raise TypeError(
                "Input frames are a mix of %(imageTypes)s" % locals())

        self.log.debug('completed the ``verify_input_frames`` method')
        return None

    def produce_product(
            self):
        """*The code to generate the product of the mdark recipe*

        **Return:**
            - ``productPath`` -- the path to the final product

        **Usage:**

        ```python
        from soxspipe.recipes import mdark
        recipe = mdark(
            log=log,
            settings=settings,
            inputFrames=fileList
        )
        mdarkFrame = recipe.produce_product()
        ```
        """
        self.log.debug('starting the ``produce_product`` method')

        # IMAGECOLLECTION FILEPATHS
        filepaths = self.inputFrames.files_filtered(include_path=True)

        productPath = None

        self.log.debug('completed the ``produce_product`` method')
        return productPath

    # use the tab-trigger below for new method
    # xt-class-method

    # Override Method Attributes
    # method-override-tmpx



#!/usr/bin/env python
# encoding: utf-8
"""
*The base recipe class which all other recipes inherit*

:Author:
    David Young & Marco Landoni

:Date Created:
    January 22, 2020
"""
################# GLOBAL IMPORTS ####################
from builtins import object
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools
import numpy as np
from astropy.nddata import CCDData
from astropy import units as u
import ccdproc
from soxspipe.commonutils import set_of_files
from soxspipe.commonutils import keyword_lookup


class _base_recipe_(object):
    """
    The base recipe class which all other recipes inherit

    **Key Arguments:**
        - ``log`` -- logger
        - ``settings`` -- the settings dictionary

    **Usage:**

    To use this base recipe to create a new `soxspipe` recipe, have a look at the code for one of the simpler receipes (e.g. `mbias`) - copy and modify the code.
    """
    # Initialisation

    def __init__(
            self,
            log,
            settings=False,

    ):
        self.log = log
        log.debug("instansiating a new '__init__' object")
        self.settings = settings
        self.intermediateRootPath = self._absolute_path(
            settings["intermediate-data-root"])
        self.reducedRootPath = self._absolute_path(
            settings["reduced-data-root"])
        self.calibrationRootPath = self._absolute_path(
            settings["calibration-data-root"])
        # xt-self-arg-tmpx

        return None

    def prepare_single_frame(
            self,
            frame,
            save=False):
        """*prepare a single raw frame by converting to electron counts and adding mask and uncertainty extensions*

        **Key Arguments:**
            - ``frame`` -- the path to the frame to prepare, of a CCDData object

        **Return:**
            - ``frame`` -- the prepared frame with mask and uncertainty extensions (CCDData object)

        **Usage:**

        ```python
        from soxspipe.recipes import _base_recipe_
        recipe = _base_recipe_(
            log=log,
            settings=settings
        )
        preFrame = recipe.prepare_single_frame(frame="/path/to/frame.fits")
        ```

        ---

        ```eval_rst
        .. todo::

            - write a command-line tool for this method
        ```

        # Preparing the Raw SOXS Frames

        # Trim Overscan

        The first thing we need to do is trim off the overscan area of the image.

        # ADU to Electrons

        The first thing to do is to convert the pixel data from ADU to electron counts by multiplying each pixel value in the raw frame by the gain (`HIERARCH ESO DET OUT1 CONAD` keyword value).

        $$\rm{electron\ count} = \rm{adu\ count} \times \rm{gain}$$

        # Generating an Uncertainty Map

        Next we need to generate the uncertainty map for the raw image and add this as the 'UNCERT' extension to the image.

        For each pixel the uncertainty is calculated as:

        $$\rm{error} = \sqrt{\rm{readnoise}^2+\rm{electron\ count}}$$

        # Bitmap Extension

        The appropriate bitmap extension is selected and simply added as the 'FLAG' extension of the frame.

        # Bad Pixel Mask

        The selected bitmap is converted to a boolean mask, with values >0 becoming TRUE to indicate these pixels need to be masks. All other values are set to FALSE.

        Finally the prepared frames are saved out into the intermediate frames location with the prefix `pre_`.

        Viewing the image in DS9 (using the command `ds9 -multiframe pre_filename.fits` to show all extensions as tiled frames) we can see the 'DATA', 'UNCERT' and 'MASK' extensions are now all present.

        ![](https://live.staticflickr.com/65535/49434143241_580e904616_o.png)
        """
        self.log.debug('starting the ``prepare_single_frame`` method')

        # KEYWORD LOOKUP OBJECT - LOOKUP KEYWORD FROM DICTIONARY IN RESOURCES
        # FOLDER
        kw = keyword_lookup(
            log=self.log,
            settings=self.settings
        ).get

        # CONVERT FILEPATH TO CCDDATA OBJECT
        filepath = frame
        if isinstance(frame, str):

            # CONVERT RELATIVE TO ABSOLUTE PATHS
            frame = self._absolute_path(frame)

            # OPEN THE RAW FRAME
            frame = CCDData.read(frame, hdu=0, unit='electron', hdu_uncertainty='UNCERT',
                                 hdu_mask='MASK', hdu_flags='BITMAP', key_uncertainty_type='UTYPE')

        # CHECK THE NUMBER OF EXTENSIONS IS ONLY 1
        if len(frame.to_hdu()) > 1:
            raise TypeError("%(frame)s is not a raw frame" % locals())

        # TRIM OVERSCAN REGION
        if frame.header[kw('SEQ_ARM')] != "NIR":
            frame = ccdproc.trim_image(frame[:, :2048])
        else:
            frame = ccdproc.trim_image(frame[:1056, :2040])

        # HIERARCH ESO DET OUT1 CONAD - Electrons/ADU
        if (kw('DET_GAIN') not in frame.header) and (kw('DET_CHIP_GAIN') not in frame.header):
            one = kw('DET_GAIN')
            two = kw('DET_CHIP_GAIN')
            raise AttributeError(
                "'%(one)s/%(two)s' keyword not found in %(frame)s" % locals())
        if kw('DET_GAIN') in frame.header:
            gain = frame.header[kw('DET_GAIN')]
        else:
            gain = frame.header[kw('DET_CHIP_GAIN')
                                ]

        # HIERARCH ESO DET OUT1 RON - Readout noise in electrons
        if (kw('RON') not in frame.header) and (kw('CHIP_RON') not in frame.header):
            one = kw('RON')
            two = kw('CHIP RON')
            raise AttributeError(
                "'%(one)s/%(two)s' keyword not found in %(frame)s" % locals())
        if kw('RON') in frame.header:
            ron = frame.header[kw('RON')]
        else:
            ron = frame.header[kw('CHIP_RON')]

        # CONVERT ADU TO ELECTRONS
        frame.data = frame.data * gain

        # GENERATE UNCERTAINTY MAP AS EXTENSION
        frame = ccdproc.create_deviation(
            frame, readnoise=ron * u.electron)

        # FIND THE APPROPRIATE BAD-PIXEL BITMAP AND APPEND AS 'FLAG' EXTENSION
        # NOTE FLAGS NOTE YET SUPPORTED BY CCDPROC THIS THIS WON'T GET SAVED OUT
        # AS AN EXTENSION
        arm = frame.header[kw('SEQ_ARM')]
        if "NIR" in arm:
            bitMapPath = self.calibrationRootPath + \
                "/cal/BP_MAP_RP_%(arm)s.fits" % locals()
        else:
            binx = int(frame.header[kw('WIN_BINX')])
            biny = int(frame.header[kw('WIN_BINY')])
            bitMapPath = self.calibrationRootPath + \
                "/cal/BP_MAP_RP_%(arm)s_%(binx)sx%(biny)s.fits" % locals()

        bitMap = CCDData.read(bitMapPath, hdu=0, unit='electron')
        if "NIR" in arm:
            bitMap.data = np.rot90(bitMap.data, -1)
        frame.flags = bitMap.data

        # FLATTEN BAD-PIXEL BITMAP TO BOOLEAN FALSE (GOOD) OR TRUE (BAD) AND
        # APPEND AS 'UNCERT' EXTENSION
        boolMask = bitMap.data.astype(bool).data
        try:
            # FAILS IN PYTHON 2.7 AS BOOLMASK IS A BUFFER - NEED TO CONVERT TO
            # 2D ARRAY
            boolMask.shape

        except:
            arr = np.frombuffer(boolMask, dtype=np.uint8)
            arr.shape = (frame.data.shape)
            boolMask = arr

        frame.mask = boolMask

        if save:
            outDir = self.intermediateRootPath
        else:
            outDir = self.intermediateRootPath + "/tmp"

        # RECURSIVELY CREATE MISSING DIRECTORIES
        if not os.path.exists(outDir):
            os.makedirs(outDir)
        # CONVERT CCDData TO FITS HDU (INCLUDING HEADER) AND SAVE WITH PRE TAG
        # PREPENDED TO FILENAME
        basename = os.path.basename(filepath)
        filenameNoExtension = os.path.splitext(basename)[0]
        extension = os.path.splitext(basename)[1]
        filePath = outDir + "/" + \
            filenameNoExtension + "_pre" + extension
        frame.write(filePath, overwrite=True)

        self.log.debug('completed the ``prepare_single_frame`` method')
        return filePath

    def _absolute_path(
            self,
            path):
        """*convert paths from home directories to absolute paths*

        **Key Arguments:**
            - ``path`` -- path possibly relative to home directory

        **Return:**
            - ``absolutePath`` -- absolute path

        **Usage:**

        ```python
        myPath = self._absolute_path(myPath)
        ```
        """
        self.log.debug('starting the ``_absolute_path`` method')

        from os.path import expanduser
        home = expanduser("~")
        if path[0] == "~":
            path = home + "/" + path[1:]

        self.log.debug('completed the ``_absolute_path`` method')
        return path.replace("//", "/")

    def prepare_frames(
            self,
            save=False):
        """*prepare all frames in the input data*

        See `prepare_single_frame` for details.

        **Key Arguments:**
            - ``save`` -- save out the prepared frame to the intermediate products directory. Default False.

        **Return:**
            - ``preframes`` -- the new image collection containing the prepared frames

        **Usage:**

        Usually called within a recipe class once the input frames have been selected and verified (see `mbias` code for example): 

        ```python
        self.inputFrames = self.prepare_frames(
            save=self.settings["save-intermediate-products"])
        ```
        """
        self.log.debug('starting the ``prepare_frames`` method')

        # KEYWORD LOOKUP OBJECT - LOOKUP KEYWORD FROM DICTIONARY IN RESOURCES
        # FOLDER
        kw = keyword_lookup(
            log=self.log,
            settings=self.settings
        ).get

        filepaths = self.inputFrames.files_filtered(include_path=True)

        frameCount = len(filepaths)
        print("# PREPARING %(frameCount)s RAW FRAMES - CONVERTING TO ELECTRON COUNTS, GENERATING UNCERTAINTY MAPS AND APPENDING DEFAULT BAD-PIXEL MASK" % locals())
        preframes = []
        preframes[:] = [self.prepare_single_frame(
            frame=frame, save=save) for frame in filepaths]
        sof = set_of_files(
            log=self.log,
            settings=self.settings,
            inputFrames=preframes
        )
        preframes = sof.get()

        self.inputFrames.sort([kw('MJDOBS').lower()])

        self.log.debug('completed the ``prepare_frames`` method')
        return preframes

    # use the tab-trigger below for new method
    # xt-class-method



#!/usr/bin/env python
# encoding: utf-8
"""
*The recipe for creating master-bias frames *

:Author:
    David Young & Marco Landoni

:Date Created:
    January 22, 2020
"""
################# GLOBAL IMPORTS ####################
from builtins import object
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools
from soxspipe.commonutils import set_of_files
from ._base_recipe_ import _base_recipe_
from astropy.stats import mad_std
import numpy as np
from astropy.nddata import CCDData
import ccdproc
from ccdproc import Combiner


class mbias(_base_recipe_):
    """
    *The mbias recipe*

    **Key Arguments:**
        - ``log`` -- logger
        - ``settings`` -- the settings dictionary
        - ``inputFrames`` -- input fits frames. Can be a directory, a set-of-files (SOF) file or a list of fits frame paths. Default []

    **Usage:**

    To setup your logger, settings and database connections, please use the ``fundamentals`` package (`see tutorial here <http://fundamentals.readthedocs.io/en/latest/#tutorial>`_).

    See `produce_product` method for usage.


    ```eval_rst
    .. todo::

        - add a tutorial about ``mbias`` to documentation
    ```
    """
    # Initialisation

    def __init__(
            self,
            log,
            settings=False,
            inputFrames=[]

    ):
        # INHERIT INITIALISATION FROM  _base_recipe_
        super(mbias, self).__init__(log=log, settings=settings)
        self.log = log
        log.debug("instansiating a new 'mbias' object")
        self.settings = settings
        self.inputFrames = inputFrames
        # xt-self-arg-tmpx

        # INITIAL ACTIONS
        # CONVERT INPUT FILES TO A CCDPROC IMAGE COLLECTION (inputFrames >
        # imagefilecollection)
        sof = set_of_files(
            log=self.log,
            settings=self.settings,
            inputFrames=self.inputFrames
        )
        self.inputFrames = sof.get()

        # VERIFY THE FRAMES ARE THE ONES EXPECTED BY MBIAS - NO MORE, NO LESS.
        # PRINT SUMMARY OF FILES.
        print("# VERIFYING INPUT FRAMES")
        self.verify_input_frames()
        sys.stdout.write("\x1b[1A\x1b[2K")
        print("# VERIFYING INPUT FRAMES - ALL GOOD")

        print("# RAW INPUT BIAS FRAMES - SUMMARY")
        # SORT IMAGE COLLECTION
        self.inputFrames.sort(['mjd-obs'])
        print(self.inputFrames.summary, "\n")

        # PREPARE THE FRAMES - CONVERT TO ELECTRONS, ADD UNCERTAINTY AND MASK
        # EXTENSIONS
        self.inputFrames = self.prepare_frames(
            save=self.settings["save-intermediate-products"])

        return None

    def verify_input_frames(
            self):
        """*verify the input frame match those required by the mbias recipe*

        **Return:**
            - ``None``

        If the fits files conform to required input for the recipe everything will pass silently, otherwise an exception shall be raised.
        """
        self.log.debug('starting the ``verify_input_frames`` method')

        imageTypes = self.inputFrames.values(
            keyword='eso dpr type', unique=True)

        # MIXED INPUT IMAGE TYPES ARE BAD
        if len(imageTypes) > 1:
            imageTypes = " and ".join(imageTypes)
            print(self.inputFrames.summary)
            raise TypeError(
                "Input frames are a mix of %(imageTypes)s" % locals())
        # NON-BIAS INPUT IMAGE TYPES ARE BAD
        elif imageTypes[0] != 'BIAS':
            print(self.inputFrames.summary)
            raise TypeError(
                "Input frames not BIAS frames" % locals())

        arms = self.inputFrames.values(
            keyword='eso seq arm', unique=True)
        # MIXED INPUT ARMS ARE BAD
        if len(arms) > 1:
            arms = " and ".join(arms)
            print(self.inputFrames.summary)
            raise TypeError(
                "Input frames are a mix of %(imageTypes)s" % locals())

        cdelt1 = self.inputFrames.values(
            keyword='cdelt1', unique=True)
        cdelt2 = self.inputFrames.values(
            keyword='cdelt2', unique=True)
        # MIXED BINNING IS BAD
        if len(cdelt1) > 1 or len(cdelt2) > 1:
            print(self.inputFrames.summary)
            raise TypeError(
                "Input frames are a mix of binnings" % locals())

        self.log.debug('completed the ``verify_input_frames`` method')
        return None

    def produce_product(
            self):
        """*The code to generate the product of the mbias recipe*

        **Return:**
            - ``productPath`` -- the path to the final product

        **Usage:**

        ```python
        from soxspipe.recipes import mbias
        recipe = mbias(
            log=log,
            settings=settings,
            inputFrames=fileList
        )
        mbiasFrame = recipe.produce_product()
        ```
        """
        self.log.debug('starting the ``produce_product`` method')

        # IMAGECOLLECTION FILEPATHS
        filepaths = self.inputFrames.files_filtered(include_path=True)
        ccds = []
        firstInputheader = None
        for f in filepaths:
            ccd = CCDData.read(f, hdu=0, unit='electron', hdu_uncertainty='UNCERT',
                               hdu_mask='MASK', hdu_flags='BITMAP', key_uncertainty_type='UTYPE')
            if not firstInputheader:
                firstInputheader = ccd.header
            ccds.append(ccd)

        print("# CLIPPING PIXELS WITH EXTREME VALUES IN INDIVIDUAL FRAMES")
        # PRINT SOME INFO FOR USER
        arm = ccd.header['eso seq arm'].upper()
        badCount = ccd.mask.sum()
        totalPixels = np.size(ccd.mask)
        percent = (float(badCount) / float(totalPixels)) * 100.
        print("    The basic bad-pixel mask for the %(arm)s detector contains %(badCount)s pixels (%(percent)0.2f%% of all pixels)" % locals())

        # CREATE A COMBINER OBJECT
        # https://ccdproc.readthedocs.io/en/latest/api/ccdproc.Combiner.html#ccdproc.Combiner
        combiner = Combiner(ccds)

        # GENERATE A MASK FOR EACH OF THE INDIVIDUAL INOUT FRAMES - USING
        # MEDIAN WITH MEDIAN ABSOLUTE DEVIATION (MAD) AS THE DEVIATION FUNCTION
        old_n_masked = 0
        new_n_masked = combiner.data_arr.mask.sum()
        # print("The basic bad-pixel mask contains %(new_n_masked)s pixels" % locals())
        iteration = 1
        while (new_n_masked > old_n_masked):
            combiner.sigma_clipping(
                low_thresh=5, high_thresh=5, func=np.ma.median, dev_func=mad_std)
            old_n_masked = new_n_masked
            new_n_masked = combiner.data_arr.mask.sum()
            diff = new_n_masked - old_n_masked
            extra = ""
            if diff == 0:
                extra = " - we're done"
            print("    Clipping iteration %(iteration)s finds %(diff)s more rogue pixels in the set of input frames%(extra)s" % locals())
            iteration += 1

        # GENERATE THE COMBINED MEDIAN
        print("# MEDIAN COMBINING INPUT FRAMES - USING UPDATED BAD-PIXEL MASK")
        combined_bias_median = combiner.median_combine()
        combined_bias_median.header = firstInputheader
        combined_bias_median.header[
            "HIERARCH ESO PRO CATG"] = "MASTER_BIAS_%(arm)s" % locals()

        # CALCULATE NEW PIXELS ADDED TO MASK
        newBadCount = ccd.mask.sum()
        diff = newBadCount - badCount
        print("%(diff)s new pixels made it into the propagated bad-pixel map" % locals())

        x = combined_bias_median.header['eso det win1 binx']
        y = combined_bias_median.header['eso det win1 biny']
        productPath = self.intermediateRootPath + \
            "/master_bias_%(arm)s_%(x)sx%(y)s.fits" % locals()

        combined_bias_median.write(
            productPath,
            overwrite=True)

        self.log.debug('completed the ``produce_product`` method')
        return productPath

    # use the tab-trigger below for new method
    # xt-class-method

    # Override Method Attributes
    # method-override-tmpx
